#include	<stdio.h>
#include	<string.h>
#include	<ctype.h>
#include	<fcntl.h>
#include	<stdlib.h>
#ifdef		__MSDOS__
#include	<alloc.h>
#define	vpopen open
#define	vpstat stat
#else
extern int vpopen();
extern int vpstat();
#include	<unistd.h>
#include	<memory.h>
#endif
#define		PR	(void)printf
#define		SPR	(void)sprintf
#define NL '\n'

typedef enum DIRECTION {
	LEFT = 0,
	RIGHT = 1
} DIRECTION;

/* LOC: structure to define extent of a match to text. */
typedef struct LOC {
	CHAR	*loc1;	/* Pointer to first character of a match. */
	CHAR	*loc2;	/* Pointer to the last character of a match. */
} LOC;

/* CGWINDOW: struct to define extent of a match to text and the window around it. */
typedef struct CGWINDOW {
	CHAR	*wbegin; /* first character of window */
	CHAR	*begin; /* first character of match */
	CHAR	*end;   /* last character of match */
	CHAR	*wend;   /* last character of window */
	int	prev_matches;	/* # matches to end delimiter in previous portions
				 * of the window.  May be non-zero only if the
				 * window has been split to fit in buffer.
				 */
} CGWINDOW;

typedef struct PATTERN {
/* Determined in initialization or conversion */
	CHAR *cmap;	/* Character map (e.g., implements -i). */
	BOOL vflag;	/* -v one line at a time (and not converted to -V) */
	BOOL vfast;	/* -v by inversion (and not converted to -V) */
	BOOL backref;	/* grep backreferencing \(...\)...\1 */
	BOOL pbounds;	/* Need exact longest extent of match (for -t or -T) */
	BOOL backward;	/* Reversed direction of match (for -w delim) */
	BOOL approx;	/* Approximate matching (-A or +C 0) */
	BOOL is_delim;	/* This is a delimiter (+-w). */
	BOOL incl_match; /* Search for this delim in matched line? */
	BOOL incl_delim; /* Include this delim in window? */
	int  wcount;	/* Delimiter count needed to terminate window */
	int  N_start;	/* # of matches to initially skip: --# as skipped. */
	int  N_stop;	/* # of matches to do: --# as done. */
	unsigned int match_buffers; /* # of array slots for match storage */
	short cost_limit;/* Cost limit for approx matches (-A #) */
	short cost_sub;	/* Substitution cost for approx matches (+-C #) */
	short cost_del;	/* Deletion cost for approx matches (-C #) */
	short cost_ins;	/* Insertion cost for approx matches (+C #) */
	CHAR *special;	/* Character string for special -R conversion */
/* Determined in compilation */
	CHAR *non_alpha;/* Non-alphanum char class (for delimiting words) */
	RE_RE *re_ptr;	/* Compiled regular expression (or 0 ) */
	RE_CW *cw_ptr;	/* Compiled Commentz-Walter expression (or 0)*/
	RE_RE *count_ptr;/* Compiled reverse regular expression (or 0) */
	BOOL allbounds;	/* Need all extents of match (pattern has initial Group) */
	int  mult_lines;/* Max. # lines that can match expression, minus 1 */
	BOOL (*procfn)();/* First function for expression matching */
	BOOL (*succfn)();/* Second function for expression matching (if any) */
	BOOL (*thirdfn)();/* Third function for exp. matching (rarely used) */
	GR *groups;	/* Groups for backreferencing */
/* Variables */
	BOOL done;	/* Match count exhausted */
	int  nmatches;	/* # of matches stored in arrays. */
	CGWINDOW *matches;/* Window array for match storage */
	LOC  loc;	/* Useful temporary text pointers. */
/* used only in conversion or compilation */
	PARSE_TYPE parser;/* Parser used at the moment (e.g. -E -F -G) */
	CHAR *expression;/* Beginning of expression (raw or converted) */
	CHAR *exp_end;	/* End of expression (raw or converted) */
/* used only in conversion */
	CHAR *bmap;	/* Character map used for -a */
	int xflag;	/* Indicator for -x or -x1 */
} PATTERN;

#define MAX(a,b)	((a) >= (b) ? (a) : (b))
#define MIN(a,b)	((a) >= (b) ? (b) : (a))

#define CCL_SIZ		32
#define CCL_SET(a, c)	((a)[(c) >> 3] |= bittab[(c) & 07])
#define CCL_CLR(a, c)	((a)[(c) >> 3] &= ~bittab[(c) & 07])
#define CCL_CHK(a, c)	(((a)[(c) >> 3] & bittab[(c) & 07]) ? 1 : 0)

#define	errmsg0(msg)	(void)fprintf(stderr, msg, Progname)
#define	errmsg(msg, arg) (void)fprintf(stderr, msg, Progname, arg)
#define	errmsg2(msg, arg, arg2)	(void)fprintf(stderr, msg, Progname, arg, arg2)
#define	exitmsg0(msg)	{errmsg0(msg); myexit(2);}
#define	exitmsg(msg, arg) {errmsg(msg, arg); myexit(2);}
#define	exitmsg2(msg, arg, arg2) {errmsg2(msg, arg, arg2); myexit(2);}

#define FREE(p) {free((char *)(p)); p = 0;}

#define ESIZE	256
#define	BLKSIZE	512
#define	HUGE	0x10000000

#ifdef		__MSDOS__
#define MAXBUFSIZE (31*1024)
#else
#define MAXBUFSIZE (64*BUFSIZ)
#define MINBUFSIZE (1024)
#endif

#ifdef	__MSDOS__
#define LARGE	(Maxbufsize+ESIZE)
#else
#define LARGE	(HUGE+ESIZE)
#endif

/* Identifiers of patterns to be matched */
typedef enum EXP_ID {
	MAT_PATTERN = 0,	/* Pattern to be matched. */
	BEG_DELIM = 1,		/* Delimiter to begin window (-w ...) */
	END_DELIM = 2,		/* Delimiter to end window (+w ...) */
	TRAIL_PATTERN = 3,	/* Trail pattern (e.g. -t ...) */
	COMB_PATTERN = 4,	/* For fixed strings that matched trail pattern */
	NEXP_IDS	/* Must be last */
} EXP_ID;

extern CHAR	bittab[];
extern CHAR	*Buffer, *Refill_mark, *Mline_mark, *Last_NL, *Buffend;
extern CHAR	*Boffset;
extern char	*Filename, *Delimiter, *Progname;
extern BOOL	File_more;
extern BOOL	Window_truncated, Check_truncated;
extern BOOL	Line_truncated, Mline_truncated;
extern int	Maxbufsize;
extern long	Nlines[];
extern long	Mlines;

extern BOOL	NL_search, L_done, M_done;
extern int	Flags[];
extern int	Margc;
extern char	**Margv;

extern int	Nsucc;

extern long	Tln, Line_number;
extern CHAR	*Last_printed, *Last_counted;
extern BOOL	Any_printed;

extern CHAR	*window_limit();
extern void	exec_file(), match_alloc();
extern void	do_output(), compile(), convert(), windows();
extern void	trail_patterns();
extern void	myexit();
extern BOOL	grepmatch();
extern int	pattern_match();
extern void	re_error();
