	static char sccsid[]="%Z%	cgrep.c	%E% %U%";
/******************************************************************************

                ###############################################
                # Copyright 1987-2001 Lucent Technologies Inc #
                #              All Rights Reserved            #
		# Author:  Bill Tanenbaum		      #
                ###############################################

Module:	cgrep.c [%P%]

Description:

	Contains main program for "cgrep", and routines used for
	initialization and exiting due to errors.

Functions:
	main()
	fill_in()
	getoptions()
	re_error()

*******************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

/*
 * cgrep -- print lines matching (or not matching) a pattern
 *
 * Written	12/07/87
 * release 2	12/20/87
 * release 3	06/01/88
 * release 4	02/27/89
 * release 5	01/29/90
 * release 6	02/15/91
 * release 7	07/25/92
 * release 8	08/18/97
 *
 * Preprocessor variable __MSDOS__ must be set during compilation for
 * compatibility with TURBO C under MS/DOS.  TURBO C should set this automatically.
 *
 *	status returns:
 *		0 - ok, and some matches
 *		1 - ok, but no matches
 *		2 - some error
 */

/* Version is release.point_load.
 * The release is incremented for new features.
 * The point_load is incremented for bug fixes or other internal improvements.
 */
#define VERSION  "8.15"

#ifndef __MSDOS__
#include <sys/types.h>
#endif
#include <sys/stat.h>

#include "re.h"

/* Global parameters */

BOOL	Any_printed;	/* Has any output been printed? */
CHAR	*Boffset;	/* Pointer to beginning of file in text buffer
			 * (used only for offsets for -b and -b1 options).
			 * If the beginning of the file is no longer
			 * in the text buffer, Boffset will point to a
			 * location before the beginning of the buffer!
			 */
CHAR	*Buffend;	/* Pointer to last character in the text buffer. */
CHAR	*Buffer;	/* Pointer to first character in the text buffer. */
BOOL	Check_truncated;/* Does running off the beginning of the buffer in
			 * a delimiter search mean that output has been
			 * truncated?
			 */
char	*Delimiter;	/* Separator string output between non-adjacent windows. */
char	*Filename;	/* Filename of a text file */
BOOL	File_more;	/* Does more of the file need to be read in? */
int	Flags[128];	/* Flags for various command line options.  The subscript
			 * is the ASCII option letter.
			 * Flags[] is used for DHMRVabchilnostvx option letters.
			 * Flags[] is not used for ABCEFGINWdefw option letters,
			 * even though the options exist.
			 * -T uses 'R' and 't', but not 'T'.
			 */
BOOL	L_done;		/* Are we no longer looking for matches, because -l or -l1
			 * was specified, and we found a match?
			 */
CHAR	*Last_counted;	/* Pointer to character (in text buffer) that was
			 * most recently printed or counted.
			 */
CHAR	*Last_NL;	/* Pointer to last new-line character read into the
			 * text buffer.  However, if the end of the file has
			 * been read, and the last line is not terminated by
			 * a new-line, Last_NL will point to the sentinel new-line
			 * character after the last line of text.
			 */
CHAR	*Last_printed;	/* Pointer to character (in text buffer) that was
			 * most recently printed.
			 */
long	Line_number;	/* Line number of current line in file.  Not maintained
			 * unless -n was specified.
			 */
BOOL	M_done;		/* Are we no longer looking for matches for any reason
			 * (i.e. -l, -l1 or -N options).
			 */
int	Margc;		/* File count for -M (merge files) option. */
char	**Margv;	/* Array of file names for -M option */
int	Maxbufsize = MAXBUFSIZE;	/* Maximum text buffer size */
CHAR	*Mline_mark;	/* Pointer to a new-line character near the end of the
			 * text buffer.  If "Match_pattern" must fit
			 * on a single line, Mline_mark is equal to "Last_NL".
			 * Otherwise, Mline_mark points to the new-line
			 * character "Mlines" lines before "Last_NL".
			 * See also "Last_NL" and "Mlines".
			 */
long	Mlines;		/* The maximum number of new-line characters that a single
			 * instance of the text matching "Match_pattern" may contain,
			 * not including the first and last characters of the
			 * match.  (e.g. Mlines == 0, if the pattern must be on a
			 * single line).
			 */
BOOL	NL_search;	/* If a line of text needs to be output, does any other
			 * information have to be output with it
			 * (e.g. file name, line number, etc.).
			 */
long	Nlines[2];	/* Stores the arguments of the -N and +N options. */
int	Nsucc;		/* 2, if an error has occurred, otherwise 1 if a match
			 * has occurred, and 0 otherwise.  NOTE: The -s1 and
			 * -V options modify this somewhat.
			 */
char	*Progname;	/* Program name.  Used for error messages. */
CHAR	*Refill_mark;	/* If a file does not fit into the text buffer all at once,
			 * "Refill_mark" points to the beginning of the
			 * portion of the text in the buffer that will be kept
			 * in the buffer for processing on the next pass.
			 */
long	Tln;		/* Used with -c or -c1 options.  Count of lines that
			 * will be output.  Equal to the number of lines of
			 * text that would have been printed had -c or -c1 not
			 * been specified.
			 */
BOOL	Window_truncated = NO;	/* Has a window possibly been lost or truncated
				 * because the buffer size is too small?
				 */

BOOL	Line_truncated = NO;	/* Has a line possibly been lost or truncated
				 * because the buffer size is too small?
				 */

BOOL	Mline_truncated = NO;	/* Has a match possibly been missed because the
				 * buffer size is too small to hold enough
				 * lines to find a multiline match?
				 */

/* Static parameters */
static BOOL	approx = NO;	/* Is approximate matching being used, i.e.
				 * are any of "cost_del", "cost_ins" or
				 * "cost_sub" less than or equal to
				 * "cost_limit".
				 */
static CHAR	char_map[5][256];/* Character maps.  Used to map equivalent
				  * characters into one character (helps
				  * performance).  Also implements -i and -a
				  * options.
				  */
				
static short	cost_del = 1;	/* Cost of a deletion error.  (argument of -C)
				 * Must be positive integer.
				 */
static short	cost_ins = 1;	/* Cost of an insertion error. (argument of +C)
				 * Must be positive integer or zero.
				 */
static short	cost_limit = 0; /* Maximum error cost allowed for approximate
				 * matching. (argument of -A).
				 */
static short	cost_sub = 1;	/* Cost of a substitution error. (argument of +-C)
				 * Must be positive integer.
				 */
static	char	*opt_arg;	/* Temporary for argument of command line option */
static 	int	opt_ind = 1;	/* Index into argv[] of argument being processed. */
static	char	optsign;	/* Argument prefix: '-', '+', or '0'.  '0' stands
				 * for "+-" or "-+".
				 */
static PARSE_TYPE parser = grepparse;	/* -E (egrep), -F (fgrep), or -G(grep)
					 * mode
					 */
static int	r_reg;		/* Temporary for optional numeric argument for
				 * option letters that do not otherwise take
				 * an argument (e.g. -l1, -I2, -c1, etc.).
				 */


static PATTERN pattern[NEXP_IDS]; 	/* Array of PATTERN structures for patterns,
					 * window delimiters, and trail patterns
					 * specified.
					 */

/* Pointers into "pattern" for window delimiters */
static PATTERN *Begin_delim	= &pattern[(int)BEG_DELIM];
static PATTERN *End_delim	= &pattern[(int)END_DELIM];

/* Pointer into "pattern" for patterns */
static PATTERN *Match_pattern	= &pattern[(int)MAT_PATTERN];

/* Pointer into "pattern" for trail patterns */
static PATTERN *Trail_pattern	= &pattern[(int)TRAIL_PATTERN];

/* Pointer into "pattern" for union of patterns and trail patterns */
static PATTERN *Comb_pattern	= &pattern[(int)COMB_PATTERN];


static int getoption();
static void fill_in();

/******************************************************************************

Routine:

	main()

Syntax:
	main(argc, argv)

	int	argc;		// Number of command line arguments
	char	*argv[];	// Array of command line arguments

Description:

	The main routine for cgrep.

Global Variables:

	See above

Files:

	As specified in command line arguments

Return Status:

	0	// matches found
	1	// no matches found
	2	// error occurred, whether or not matches were found

*******************************************************************************/

main(argc, argv)
register int argc;
char **argv;
{
	register int	c;
	register PATTERN *p;
	char		*ptr;
	char		*file;
	int		i;
	long		lg;
	static int	errflg;
	static BOOL	flag;
	static int	Iflag[2];
	static int	wcount[2] = {-1, -1};
	static char	*expfname;
	static int	expfile;

	/* Remove the path from the program name */
	if ((Progname = strrchr(argv[0], '/')) != 0) {
		Progname++;
	} else {
		Progname = argv[0];
	}
#ifdef __MSDOS__
	/* Remove the path from the program name */
	if ((ptr = strrchr(Progname, '\\')) != 0) {
		Progname = ++ptr;
	}
	/* Remove the extension from the program name */
	if ((ptr = strrchr(Progname, '.')) != 0) {
		*ptr = '\0';
	}
#endif
	/* Set the default parser to that for grep, fgrep or egrep */
	if (!strcmp(Progname, "grep")) {
		parser = grepparse;
	} else if (!strcmp(Progname, "egrep")) {
		parser = egrepparse;
	} else if (!strcmp(Progname, "fgrep")) {
		parser = fgrepparse;
	}

	/* Loop to process options */
	while ((c=getoption(argc, argv,
		(char *)"abcd:e:f:hilnost:vw:xA:B:C:DEFGHIMN:RTVXW:")) != EOF) {
		if (optsign == '-') {
			/* Option begins with '-' (not "-+").  This switch
			 * only handles those '-' options with no analogous
			 * '+' option.
			 */
			switch(c) {
			case 'd':
				/* -d <window_separator> */
				Delimiter = opt_arg;
				continue;
			case 'e':
				/* -e <pattern(s)> */
				if (expfname) {
					errmsg0("%s: can't use both -e and -f options\n");
					errflg++;
				}
				/* Initialize pattern(s) to be matched. */
				p = Match_pattern;
				p->expression = (CHAR *)opt_arg;

				/* Initialize attributes related to the patterns. */
				fill_in(p);
				continue;
			case 'f':
				/* -f <pattern_file> */
				/* Save pattern file name */
				expfname = opt_arg;
				/* Initialize attributes related to the patterns. */
				p = Match_pattern;
				if (p->expression) {
					errmsg0("%s: can't use both -e and -f options\n");
					errflg++;
				}
				fill_in(p);
				continue;
			case 'B':
				/* -B <buffer_size_in_kB> */
				/* Save buffer size */
				Maxbufsize = strtol(opt_arg, &ptr, 10);
				if (ptr[0] != '\0' || Maxbufsize <= 0) {
					errmsg("%s: illegal value of argument %s\n", opt_arg);
					errflg++;
				}
#ifdef __MSDOS__
				else if (Maxbufsize > 31) {
					errmsg("%s: illegal value of argument %s\n", opt_arg);
					errflg++;
				}
#endif
				/* Convert buffer size to bytes */
				Maxbufsize *= 1024;
				continue;
			case 'A':
				/* -A <errors_allowed>
				 * Approximate matching to be used.
				 */
				lg = strtol(opt_arg, &ptr, 10);
				if (ptr[0] != '\0' || lg < 0 || lg >= 32768L) {
					errmsg("%s: illegal value of argument %s\n", opt_arg);
					errflg++;
				}
				/* Save the number of errors allowed. */
				cost_limit = lg;
				/* We use approximate matching if either:
				 * a) More than zero errors are allowed.
				 * OR
				 * b) The cost of an insertion error is zero.
				 */
				approx = (BOOL)(cost_limit > 0 || cost_ins == 0);
				continue;
			case 'T':
				/* -T (also implies -R). */
				/* Set -t flag */
				Flags['t'] = 1;
				/* Set up Trail Patterns to be matched.  */
				p = Trail_pattern;
				p->expression = (CHAR *)"EVENTO?=-?[0-9]+";
				/* Use egrep mode for -T trail patterns */
				parser = egrepparse;
				/* Initialize related attributes. */
				fill_in(p);
				/* -x or -x1 do not apply to trail patterns */
				p->xflag = 0;
				/* Approximate matching does not apply to
				 * trail patterns.
				 */
				p->approx = NO;
				p->cost_limit = 0;
				p->cost_sub = p->cost_del = p->cost_ins = 1;
				/* Make copy of trail patterns and store in
				 * Comb_pattern.
				 */
				*Comb_pattern = *Trail_pattern;
				/* We must know exact bounds of match. */
				p->pbounds = YES;
				/* FALLTHROUGH to case 'R' */
			case 'R':
				/* -R (or -T). */
				/* Set -R, -M, -h, and -D flags */
				Flags['R'] = Flags['M'] = Flags['h'] = Flags['D'] = 1;
				/* Set up beginning delimiter to be matched.  */
				p = Begin_delim;
				p->expression = (CHAR *)"\\+\\+\\+|\031";
				/* Use egrep mode for delimiter */
				parser = egrepparse;
				/* Initialize related attributes. */
				fill_in(p);
				/* Approximate matching does not apply to
				 * -R option delimiters.
				 */
				p->xflag = 0;
				p->approx = NO;
				p->cost_limit = 0;
				p->cost_sub = p->cost_del = p->cost_ins = 1;
				/* Copy beginning delimiter and its related
				 * attributes into the end delimiter.
				 * They are the same.
				 */
				*End_delim = *p;
				/* -I and +I2 are implied for -R delimiters. */
				Iflag[0] = 0;
				Iflag[1] = 2;
				/* "sm=" has special meaning when -R is used. */
				Match_pattern->special = (CHAR *)"sm=";
				continue;
			case 'G':
				/* -G (grep mode)
				 * Set default parser to "grep" for subsequent
				 * patterns.
				 */
				parser = grepparse;
				continue;
			case 'F':
				/* -F (fgrep mode)
				 * Set default parser to "fgrep" for subsequent
				 * patterns.
				 */
				parser = fgrepparse;
				continue;
			case 'E':
				/* -E (egrep mode)
				 * Set default parser to "egrep" for subsequent
				 * patterns.
				 */
				parser = egrepparse;
				continue;
			}
		}
		if (optsign != '0') {
			/* Option begins with '-' or '+' (not "-+" or "+-").
			 * This switch only handles those option letters that
			 * have related '-' and '+' cases, but do not have
			 * a "-+" or "+-" case.
			 */
			/* Usually, '-' turns on an option, while '+' turns it
			 * off.
			 */
			if (optsign == '-') {
				flag = YES;
			} else {
				flag = NO;
			}
			/* Find the maximum value (plus 1) of the optional numeric
			 * argument after the option letter.
			 */
			i = 1; /* max val of flags */
			switch(c) {
			case 'l':
				/* Max value 3 for -l */
				i++;
				/* FALLTHROUGH */ 
			case 'H':
			case 'b':
			case 'c':
			case 'h':
			case 's':
			case 'x':
				/* Max value 2 for -Hbchsx */
				i++;
				/* FALLTHROUGH */ 
			case 'V':
			case 'a':
			case 'i':
			case 'D':
			case 'M':
			case 'n':
			case 'o':
			case 'v':
				/* Max value 1 for DMVainov */
				/* Set option flag to one more than the
				 * numeric argument if '-' was used,
				 * and '0' if '+' was used.
				 */
				Flags[c] = ((int)flag ? r_reg + 1 : 0);
				/* If numeric value was too big, ignore it. */
				if (Flags[c] > i) {
					Flags[c] = 1;
				}
				continue;

			case 't':
				/* -t <trail_pattern> OR +t <trail_pattern> */
				/* Set -t flag.  (+t is identical to t) */
				Flags['t'] = 1;
				/* Set up Trail Patterns to be matched.  */
				p = Trail_pattern;
				p->expression = (CHAR *)opt_arg;
				/* Initialize related attributes. */
				fill_in(p);
				/* -x or -x1 do not apply to trail patterns */
				p->xflag = 0;
				/* Approximate matching does not apply to
				 * trail patterns.
				 */
				p->approx = NO;
				p->cost_limit = 0;
				p->cost_sub = p->cost_del = p->cost_ins = 1;
				/* Make copy of trail patterns and store in
				 * Comb_pattern.
				 */
				*Comb_pattern = *Trail_pattern;
				/* We must know exact bounds of match. */
				p->pbounds = YES;
				continue;
			}
		}
		/* Process -C, +C, +-C, or -+C */
		switch (c) {
		case 'C':
			/* -C <cost>, +C <cost>, -+C <cost>, or +-C <cost> */
			lg = strtol(opt_arg, &ptr, 10);
			if (ptr[0] != '\0' || lg < 0 || lg >= 32768L) {
				errmsg("%s: illegal value of argument %s\n", opt_arg);
				errflg++;
			}
			i = lg;
			switch(optsign) {
			case '-':
				/* -C <deletion_cost> */
				if (i == 0) {
					errmsg0("%s: Deletion cost cannot be zero.\n");
					errflg++;
				}
				cost_del = i;
				break;
			case '+':
				/* +C <insertion_cost> */
				cost_ins = i;
				/* We use approximate matching if either:
				 * a) More than zero errors are allowed.
				 * OR
				 * b) The cost of an insertion error is zero.
				 */
				approx = (BOOL)(cost_limit > 0 || cost_ins == 0);
				break;
			case '0':
				/* +-C or -+C <substitution_cost> */
				if (i == 0) {
					errmsg0("%s: Substitution cost cannot be zero.\n");
					errflg++;
				}
				cost_sub = i;
				break;
			}
			continue;
		}
		/* Process windowing options. If '-' was used, set options for
		 * beginning  delimiter.  If '+' was used, set options for
		 * end delimiter.  If "+-" or "-+" was used, set options
		 * for both delimiters.
		 * The +-N options are processed similarly, even though
		 * they have nothing to do with windowing or the delimiters.
		 */
		p = ((optsign == '+') ? End_delim : Begin_delim);
		for (; p <= ((optsign == '-') ? Begin_delim : End_delim); p++) {
			int indx = p - Begin_delim;
			/* indx is 0 for begin delimiter,
			 * and 1 for end delimiter.
			 */ 
			switch (c) {
			case 'I':
				/* {-|+|-+|+-}I[1|2].  Set Iflag[indx]
				 * to 0 for I, to 1 for I1, and to 2 for I2.
				 */
				Iflag[indx] = 0 + r_reg;
				break;
			case 'X':
				/* X is equivalent to I2. */
				Iflag[indx] = 2;
				break;
			case 'W':
			case 'w':
				/* {-|+|-+|+-}{w|W} <delimiter_pattern(s)>. */
				/* Set "vflag" if 'W' was used (in that
				 * case, delimiter is any line that
				 * does NOT match the delimiter patterns).
				 */
				p->vflag = (BOOL)(c == 'W');
				/* Fill in delimiter pattern(s) and related
				 * parameters.
				 */
				p->expression = (CHAR *)opt_arg;
				fill_in(p);
				break;
			case '0':
				/* {-|+|-+|+-} <# of lines or delimiters>. */
				/* windowing has been specified by a line
				 * count (the default) or a delimiter count
				 * (if corresponding w or W option is also
				 * specified).
				 */
				i = strtol(opt_arg, &ptr, 10);
				if (ptr[0] != '\0' || i < 0) {
					errmsg("%s: illegal character[s] in argument %s\n", opt_arg);
					errflg++;
				}
				/* Save the count. */
				wcount[indx] = i;
				break;
			case 'N':
				/* -N <max # of matches> 
				 * OR
				 * +N <# of matches to be skipped>
				 *
				 * Just store in Nlines[0] or Nlines[1] respectively.
				 */
				i = strtol(opt_arg, &ptr, 10);
				if (ptr[0] != '\0' || i < 0 ) {
					errmsg("%s: illegal character[s] in argument %s\n", opt_arg);
					errflg++;
				}
				Nlines[indx] = i;
				break;
			case '?':
				/* Illegal option. Error message already given
				 * in getoption().
				 */
				errflg++;
				break;
			default:
				/* Illegal option */
				switch (optsign) {
				case '-':
				case '+':
					errmsg2("%s: illegal option %c%c\n", optsign, c);
					break;
				default:
					errmsg2("%s: illegal option +-%c or -+%c\n", c, c);
					break;
				}
				errflg++;
				break;
			}
			if (errflg) {
				break;
			}
		}
	}

	/* If neither -e or -f was specified, the pattern(s) to be matched
	 * must be the next argument.  In that case, initialize the patterns
	 * and related parameters.
	 */
	p = Match_pattern;
	if (!p->expression && !expfname) {
		p->expression = (CHAR *)argv[opt_ind++];
		fill_in(p);
	}

	/* If any errors occurred, or we have a missing argument, give
	 * the cgrep version number, the usage message, and exit with an error.
	 */
	if (errflg || (opt_ind > argc)) {
		(void)fprintf(stdout, "This is cgrep version %s\n", VERSION);
		exitmsg0("Usage: %s [ -E | -F | -G ] [ {-|+}ainovDMV ]\n\t[ {-|+}bchsxH[1] ] [ {-|+}l[1|2] ]\n\t[ -nline ] [ +nline ] [ {-+|+-}nline ]\n\t[ -{w|W} delimiters ] [ +{w|W} delimiters ]\n\t[ {-+|+-}{w|W} delimiters ]\n\t[ -{I|I1|I2} ] [ +{I|I1|I2} ] [ {-+|+-}{I|I1|I2} ]\n\t[ -N nmatches ] [ +N nmatches ] [ {-+|+-}N nmatches ]\n\t[ -d output_separator ] [ -B text_buffer_size_in_kB)] \n\t[ {-|+}t trail_patterns | -T ] [ -R ] \n\t[ -A errors_allowed ] [ {-+|+-}C substitution_cost ]\n\t[ -C deletion_cost ] [ +C insertion_cost ]\n\t{ -f patterns_file | [-e] patterns } [files]\n");
	}

	/*
	 * If -f was used, read in the patterns to be matched from a file.
	 */
	if (expfname) {
		unsigned int _i;
		long _l;

		p = Match_pattern;

		/* The file is opened using viewpathing. */
		if ((expfile = vpopen(expfname, O_RDONLY)) < 0 ||
			(_l = lseek(expfile, (off_t)0L, 2)) < 0) {
			exitmsg("%s: can't open %s\n", expfname);
		}
		/* We allocate a buffer two bytes larger than the size
		 * of the patterns file.
		 */
		_i = _l;
		p->expression = (CHAR *)cgmalloc(_i + 2, "pattern file buffer");
		(void)lseek(expfile, (off_t)0L, 0);

		/* Read in the entire file. */
		_i = read(expfile, (char *)p->expression, _i);
		/* If the file was empty, put a new-line
		 * at the beginning of the buffer
		 * and set the parser to "nullparse",
		 * so we never find any matches.
		 */
		if (_i == 0) {
			p->expression[_i++] = NL;
			p->parser = nullparse;
		}
		/* Terminate the buffer with a NUL character.
		 * If the last character in the buffer is a new line,
		 * write over a new line.  Otherwise, put the NULL after
		 * the last character.
		 */
		if (p->expression[_i - 1] == NL) {
			p->expression[_i - 1] = '\0';
		} else {
			p->expression[_i] = '\0';
		}
		/* close the file */
		(void)close(expfile);
	}

	/*
	 * Fill in the character maps.
	 * char_map[0] maps every character to itself.
	 *
	 * char_map[1] maps every character to itself, except upper case
	 * letters maps to lower case.  Used for the -i option.
	 *
	 * char_map[2] maps every character to itself, except for characters
	 * that have special meanings when used after backslash when the -a
	 * option is used.  This special purpose mapping is used only in
	 * the initial pattern conversion (convert()) when the -a option
	 * is used.
	 *
	 * char_map[3] and char_map[4] are identical to 0 and 1 except all
	 * characters other than letters, numbers, and underscore are mapped
	 * to New-Line.  Used for the -x1 option, or matching '\<' or \>'
	 * in grep mode.
	 */
	for (i = 0; i < 256; i++) {
		char_map[4][i] = char_map[1][i] = toupper(i);
		char_map[3][i] = char_map[2][i] = char_map[0][i] = (CHAR)i;
		if (!isalnum(i)) char_map[3][i] = char_map[4][i] = NL;
	}
	char_map[3]['_'] = char_map[4]['_'] = '_';
	for (i = (int)'0'; i < (int)'8'; i++) char_map[2][i] = 0;
	char_map[2][(int)'a'] = 0x7;
	char_map[2][(int)'b'] = 0x8;
	char_map[2][(int)'t'] = 0x9;
	char_map[2][(int)'n'] = 0xa;
	char_map[2][(int)'v'] = 0xb;
	char_map[2][(int)'f'] = 0xc;
	char_map[2][(int)'r'] = 0xd;

	/*
	 * Set argc and argv to apply to any remaining arguments,
	 * which should be test file names.
	 */
	argv = &argv[opt_ind];
	argc -= opt_ind;

#ifdef __MSDOS__
	/* The MS-DOS shell does not do its own wild card expansion
	 * of filenames.  We do it here.
	 */
	if (argc > 0) {
		proc_cmd_args(argc, argv);
		argc = Margc;
		argv = Margv;
		Margc = 0;
	}
#endif

	/*
	 * Some input parameters must be modified for various reasons.
	 * Do so here.
	 * Some flags become irrelevant based on the values of various
	 * parameters and conditions.  In this case, standardize the
	 * values of the irrelevant flags.
	 * Note that the order we do these is important.
	 */

	/* -V makes -o and -H# irrelevant */
	if (Flags['V']) {
		Flags['o'] = Flags['H'] = 0;
	}

	/* -H is irrelevant if stdout is not a terminal. */
	if (Flags['H'] == 1 && !isatty(1)) {
		Flags['H'] = 0;
	}

	/* -c# conflicts with -l#. By convention, -c# wins. */
	if (Flags['c']) {
		Flags['l'] = 0;
	}

	/* -c# or -l# makes -b#, -n, -H#, -D, and -d irrelevant .*/
	if (Flags['c'] || Flags['l']) {
		Flags['b'] = Flags['n'] = Flags['H'] = 0;
		Delimiter = 0;
		Flags['D'] = 1;
	}

	/* Flags ['b'] will be set to "1" if -b1 was specified, and
	 * to the disk block size if -b was specified.
	 */
	if (Flags['b']) {
		Flags['b'] = (Flags['b'] == 1 ? BLKSIZE : 1);
	}

	/* -l# makes -h# irrelevant.  */
	if (Flags['l']) {
		Flags['h'] = 1;
	}

	/* Set Flags['h'] to 1 if file names will not be printed,
	 * and 0 if file names will be printed.
	 */
	 
	switch(Flags['h']) {
	case 1: break; /* -h means names never printed. */
	case 2: Flags['h'] = 0; break; /* -h1 means names always printed. */
	default: Flags['h'] = (argc <= 1 || Flags['M']); break; /* See below */
		/* -h# unused.  File names will not be printed if either
		 * 1) We are reading from stdin (argc == 0) OR
		 * 2) There is only one file argument (argc == 1) OR
		 * 3) -M is used (concatenate files).
		 */
	}

	/* -l# without -V makes +-w, +-nlines, +-t, and -o irrelevant,
	 * and implies that we can stop after the first match
	 * (i.e., implies -N 1)
	 */
	if (Flags['l'] && !Flags['V']) {
		wcount[0] = wcount[1] = Flags ['t'] = Flags['o'] = 0;
		Begin_delim->expression = End_delim->expression =
			Trail_pattern->expression = 0;
		Nlines[0] = 1;
	}

	/* null trail pattern makes -t and +t irrelevant */
	p = Trail_pattern;
	if (!p->expression || !p->expression[0]) {
		Flags['t'] = 0;
		p->expression = 0;
	}

	/* Begin_delim is matched backwards. */
	Begin_delim->backward = YES;

	/* These apply to both delimiters, so we use a loop. */
	for (p = Begin_delim; p <= End_delim; p++) {
		int indx = p - Begin_delim;
		/* indx is 0 for begin delimiter,
		 * and 1 for end delimiter.
		 */ 

		/* explicit +-0 overrides w */
		if (!wcount[indx]) {
			p->expression = 0;
		}

		/* wcount (nline) defaults to 1 with w, 0 without w */
		if (wcount[indx] == -1) {
			wcount[indx] = (p->expression ? 1 : 0);
		}
	}

	/* Set appropriate parameters for for the window delimiters.  */

	for (p = Begin_delim; p <= End_delim; p++) {
		int indx = p - Begin_delim;

		/* Number of matches needed to delimit window. */
		p->wcount = wcount[indx];
		/* This pattern is a delimiter */
		p->is_delim = YES;

		/* Convert +- nlines to a NL delimiter */
		if (p->wcount && !p->expression) {
			/* Use new-line delimiter with {+|-}I1 implied. */
			Iflag[indx] = 1;
			p->expression = (CHAR *)"\\\n";
			/* There is a special parser and matching routine for
			 * matching every line.
			 */
			p->parser = nlparse;
			/* Approximate matching does not apply. Also -v does
			 * not apply.
			 */
			p->vflag = p->approx = NO;
			p->xflag = 0;
			p->cost_limit = 0;
			p->cost_sub = p->cost_del = p->cost_ins = 1;
			p->cmap = p->bmap = char_map[0];
		}
		/* set up the I, I1, or I2 options.
		 * "incl_delim" is set if the line containing the delimiter
		 * is to be included in the window.  It is set for I or I1, but
		 * not for I2.
		 * "incl_match" is set if the line containing the matched
		 * pattern is to be itself searched for the window delimiter.
		 * It is set for I, but not for I2.
		 */
		p->incl_delim = YES;
		p->incl_match = NO;
		switch(Iflag[indx]) {
		case 0: default: p->incl_match = YES; break;
		case 1: break;
		case 2: p->incl_delim = NO; break;
		}
		/* Allocate one slot for a match to each window delimiter.
		 * One slot is all that is needed because only the most recent
		 * match is saved.  Previous matches are overwritten.
		 */
		p->match_buffers = 1;
		p->matches = (CGWINDOW *)
		    cgmalloc(sizeof(CGWINDOW), "delim match buffer");
	}

	if (Begin_delim->wcount || End_delim->wcount) {
		/* If windowing, and neither -d or -D is in effect,
		 * we set the default separator between non-adjacent windows.
		 */
		if (!Delimiter && !Flags['D']) {
			Delimiter = "========================================";
		}
	} else {
		/* If not windowing, -o is irrelevant. */
		Flags['o'] = 0;

		/* If not windowing, and -+t or +-N is not in effect,
		 * we can convert -v <-> -V, since they produce the same
		 * output, although internally they are processed differently.
		 * We do this if it will help performance.
		 * -V is faster, unless -l# is specified, in which case, -v is
		 * faster.
		 *
		 * NOTE: If windowing, or counting matches (+-N), or using
		 * trail patterns (+-t, -T), -v and -V are NOT equivalent,
		 * and cannot be converted to each other.
		 */
		if (!Flags['t'] && !Nlines[0] && !Nlines[1]) {
			if (Flags['v'] && Flags['V']) {
				/* Both -V and -v were used.  They cancel! */
				Flags['v'] = Flags['V'] = 0;
				/* -l# without -V implies that we can stop
				 * after the first match (i.e. like -N 1).
				 */
				if (Flags['l']) {
					Nlines[0] = 1;
				}
			} else if (Flags['V'] && Flags['l']) {
				/* -V was specified, and -l# was also.  Convert
				 * to -v for better performance in this case.
				 */
				Flags['V'] = 0;
				Flags['v'] = 1;
				/* -l# without -V implies that we can stop
				 * after the first match (i.e. like -N 1).
				 */
				Nlines[0] = 1;
			} else if (Flags['v'] && !Flags['l']) {
				/* -v was specified, and -l# was not.  Convert
				 * to -V for better performance.
				 */
				Flags['V'] = 1;
				Flags['v'] = 0;
			}
		}
	}

	/* Set -v indicator.  There are three mutually exclusive ways of
	 * handling the -v option.  The choice is determined by performance
	 * based on other options used.
	 *
	 * Case 1:  When -v is equivalent to -V, and -V is faster, -v can be
	 * converted to -V.   If this case applies, it has already been done
	 * just above, and needs no further processing.  -V uses the global
	 * Flags['V'] indicator to invert output, and does not directly
	 * affect pattern matching.
	 * 
	 * Case 2: If -v was not converted to -V above, and there is no limit
	 * to the number of "matches", we set the "vfast" indicator.
	 * The pattern will be matched everywhere in the buffer in a single pass.
	 * The matches will then be "inverted" en-masse before counts
	 * and/or windowing is applied.
	 * This is a big performance improvement over line by line matching,
	 * especially if the pattern is not found very often.
	 *
	 * Case 3: If -v was not converted to -V above, and there is a limit
	 * to the number of "matches", we set the "vflag" indicator.
	 * The pattern will be searched for line by line, stopping when the maximum
	 * number of unmatched lines is reached.
	 *
	 * Note:  This choice must be made before the patterns are compiled,
	 * as "vflag" affects compilation of the patterns, although "vfast"
	 * and "Flags['V']" do not.
	 */

	p = Match_pattern;
	if (Flags['v']) {
		if (Nlines[0]) {
			p->vflag = YES;
		} else {
			p->vfast = YES;
		}
	}

	/* Do NLs have to be found for counting or printing lines?
	 * NL_search will be set if any other information needs
	 * to be printed out every line, such as file name,
	 * line number, etc.  Highlighting does not count, because
	 * it need not be set on each line.
	 */
        NL_search = (BOOL)(Flags['n'] || !Flags['h'] || Flags['c'] ||
		Flags['b']);

	/*
	 * We are done with setting and converting input parameters.
	 */

	/* Each expression to be matched is now converted to
	 * one of two internal formats and then compiled.
	 * See the listing of convert() and compile() for details.
	 */

	for (p = Match_pattern ; p <= Trail_pattern; p++) {
		/* If the expression was not specified, there is nothing
		 * to convert or compile.
		 */
		if (!p->expression) {
			continue;
		}
		/* Convert the expression. */
		convert(p);

		/* Compile the expression. */
		compile(p);

		/* 
		 * If the pattern can match no lines, we can turn off windowing,
		 * trail patterns, -o, and match_counts.  Note: We don't
		 * have to check for -v, because compile() turns off -v
		 * if the pattern matches every line or no lines.
		 */
		if (p == Match_pattern && p->procfn == re_nullexec) {
			wcount[0] = wcount[1] = Flags ['t'] = Flags['o'] =
				Nlines[0] = Nlines[1] = 0;
			Begin_delim->expression = End_delim->expression =
				Trail_pattern->expression = 0;
		}
		/* If the pattern matches every line, we may want to optimize
		 * to match no lines and either set -v or invert -V  (not both).
		 *
		 * Note: We don't have to check for -v already being set,
		 * because compile() turns off -v if the pattern matches
		 * every line or no lines.
		 *
		 * Note: If -N was used, we will stop after a fixed number
		 * of matches.  So, it is probably faster to just count every line
		 * as we match them.  Do not optimize in this case.
		 */
		if (p == Match_pattern && p->procfn == re_lnexec && !p->succfn && !Nlines[0]) {
			if (Nlines[1]) {
				/* -N was not used, but +N was.
				 * In this case, we are counting matches to skip,
				 * so -V and -v are not equivalent.  We must
				 * leave -V alone.  But, since we will never
				 * be able to stop looking for matches,
				 * it is much faster to turn -v on, and
				 * set the pattern matching to find nothing,
				 * rather than to find each match line by line.
				 */
				p->vfast = YES;
				/* Use the matching routine that matches nothing!
				 * It is really fast!
				 */
				p->procfn = re_nullexec;
			} else {
				/* Neither -N or +N was used.  Therefore,
				 * every line of the file will be matched,
				 * and no matches will be ignored.
				 * Therefore, we can turn off windowing,
				 * trail patterns, and -o, since the entire
				 * file will be matched anyway.
				 */
				wcount[0] = wcount[1] = Flags ['t'] = Flags['o'] = 0;
				Begin_delim->expression = End_delim->expression =
					Trail_pattern->expression = 0;

				/* Since we will never be able to stop looking for
				 * matches, it is faster to turn -v on, and
				 * set the pattern matching to find nothing,
				 * rather than to find each match line by line.
				 * However, since -v and -V are now
				 * effectively equivalent, we can to invert -V
				 * instead.  This is even faster!
				 */
				Flags['V'] = (Flags['V'] ? 0 : 1);
				/* Use the matching routine that matches nothing!
				 * It is really fast!
				 */
				p->procfn = re_nullexec;
			}
		}
#ifdef	DEBUG
		if (p->procfn == re_cwexec || p->procfn == re_cbexec) {
			if (p->succfn) printf("PARTIAL COMMENTZ_WALTER\n");
			else printf("FULL COMMENTZ_WALTER\n");
		} else if (p->procfn == re_lnexec || p->procfn == re_lbexec) {
			if (p->succfn) printf("SLOW \n");
			else printf("EVERY LINE\n");
		} else if (p->procfn == re_nullexec) {
			printf("NO LINES\n");
		} else {
			printf("GENERAL\n");
		}
#endif
	}

	/* Store the maximum number of lines (minus 1) that can match a
	 * pattern or a trail pattern.
	 */
	Mlines = (Flags['t'] ?
	    MAX(Match_pattern->mult_lines, Trail_pattern->mult_lines) :
	    Match_pattern->mult_lines);

	/* Allocate match buffers */
	match_alloc(Match_pattern);
#ifndef __MSDOS__
	/* We attempt to allocate a text buffer large enough to hold the
	 * largest file (or, if -M was used, to hold all the files at once),
	 * except we never try to allocate more than Maxbufsize bytes.
	 * (Note: ESIZE extra bytes are allocated as a safety factor
	 * and to hold sentinel new-line characters before and after.)
	 * If any input is not a regular file or a directory (e.g. it's
	 * a pipe or a special file), we treat it as a file of size Maxbufsize
	 * (specified with -B or defaulted).
	 */
	{
		struct stat statbuf;
		unsigned int	bufsize = 0;
		unsigned int	filesize;

		if (argc == 0) {
			/* Input is from standard input.
			 * We can still get the size of the input file
			 * if it is a regular file or directory.   However,
			 * if it is a pipe or special device file, we use
			 * the default or specified size.
			 */

			filesize = Maxbufsize;
			if (fstat(0, &statbuf) == 0 &&
			    (statbuf.st_mode & (S_IFREG | S_IFDIR))) {
				/* 
				 * PSF 7/11/06.  We've seen cases where
				 * stdin is a pipe, yet the stat() returns
				 * success with a size of zero.  That's not
				 * good, so in that case, we continue with
				 * the default buffer rather than the bogus
				 * file size.
				 */
				if ( statbuf.st_size ) {
					filesize = statbuf.st_size;
				}	
			}
			bufsize = filesize;
		} else {
			/* Input is from specified file name(s).  Get the size
			 * of the largest such file, or, if -M was specified,
			 * the sum of the sizes of all such files.
			 */
			Margc = argc;
			Margv = argv;

			while (Margc-- > 0) {
				file = *Margv++;
				filesize = Maxbufsize;
				if (vpstat(file, &statbuf) == 0 &&
				    (statbuf.st_mode & (S_IFREG | S_IFDIR))) {
					filesize = statbuf.st_size;
				}
				if (Flags['M']) {
					bufsize += filesize;
					if (bufsize > Maxbufsize) {
						break;
					}
				} else if (filesize > bufsize) {
					bufsize = filesize;
				}
			}
			Margc = 0;
		}
		/* Add a safety factor to the buffer size. */
		bufsize += ESIZE;
		/* Don't allocate more than the default or specified size. */
		if (bufsize > Maxbufsize) {
			bufsize = Maxbufsize;
		}
		/* We now attempt to allocate space for the buffer.  If the
		 * allocation fails, we cut the buffer size in half and try
		 * again, until we succeed, or we go below MINBUFSIZE.
		 * We reset Maxbufsize to the successfully allocated size.
		 */
		while (bufsize > MINBUFSIZE) {
			Buffer = (CHAR *)malloc(bufsize + ESIZE);
			if (Buffer) {
				Maxbufsize = bufsize;
#ifdef DEBUG
	printf("malloc %d for window Bffr\n", bufsize+ESIZE);
#endif
				break;
			}
			bufsize /= 2 ;
		}
		/* If all our allocation attempts fail, we will try to
		 * allocate MINBUFSIZE bytes as a last try.
		 */
		if (!Buffer) Maxbufsize = MINBUFSIZE;
	}
#endif

	if (!Buffer) {
		/* 
		 * Allocate space for the text buffer, if not yet done so.
		 */
		Buffer = (CHAR *)cgmalloc((unsigned int)Maxbufsize + ESIZE, "window Bffr");
	}

	*Buffer = NL;  /* New line precedes buffer to prevent runover */

	/* Now we call exec_file() for each input file to do the real work.
	 * If standard input or -M was used, exec_file() is called only once.
	 */
	if (argc == 0) {
		Filename = "<stdin>";
		file = 0;
		exec_file(file, Match_pattern, Begin_delim, End_delim, Trail_pattern, Comb_pattern);
	} else if (Flags['M'] && argc > 1) {
		/* -M was used with multiple files.  One call to exec_file()
		 * will process all the files.  The globals Margc and Margv
		 * are used to pass in the number of files and their names.
		 */
		Filename = "<merged_files>";
		Margc = argc;
		Margv = argv;
		Margc--;
		file = *Margv++;
		exec_file(file, Match_pattern, Begin_delim, End_delim, Trail_pattern, Comb_pattern);
	} else {
		/* Loop over text files.  exec_file() processes
		 * one file at a time.
		 */
		while (argc-- > 0) {
			Filename = file = *argv++;
			exec_file(file, Match_pattern, Begin_delim, End_delim, Trail_pattern, Comb_pattern);
		}
	}
	/* We are all done.  Issue a warning message if indicated.  The
	 * messages are self-documenting.
	 */
	if (Line_truncated == YES) {
		(void)fprintf(stderr, "cgrep: WARNING: A line may have been lost or truncated because the text buffer\n\twas too small to hold the entire line.  You are probably searching\n\ta binary file.  If you want to search a binary file, try specifying\n\ta larger buffer size with the -B option.\n");
	} else if (Mline_truncated == YES) {
		(void)fprintf(stderr, "cgrep: WARNING: A match may have been missed because the text buffer\n\twas too small to hold enough lines to find all matches\n\tto the specified multiline pattern.  You are probably searching\n\ta binary file.  If you want to search a binary file, try specifying\n\ta larger buffer size with the -B option.\n");
	}
	if (Window_truncated == YES) {
		(void)fprintf(stderr, "cgrep: WARNING: A text window may have been lost or truncated\n\tbecause the text buffer was too small.\nTry specifying a larger buffer size with the -B option.\n");
	}
	/* If an error occurred (Nsucc == 2), return 2.
	 * Otherwise, if a match was found, (Nsucc == 1), return 0.
	 * Otherwise, if no match was found, (Nsucc == 0), return 1.
	 */
	return(Nsucc == 2 ? 2 : Nsucc == 0);
}

/******************************************************************************

Routine:
	fill_in()

Syntax:
	static void fill_in(p)

	PATTERN *p;	// Pointer to PATTERN structure to be filled in

Description:

	Fills in some properties of the pattern, based on command line input
	or defaults.  The pattern may be the match pattern, a window
	delimiter, or a trail pattern.

	This routine fills in the properties specified by
	command line options that are positional, i.e., they affect all
	patterns following them on the command line, but not those preceding
	them on the command line.  These options are:

	-E, -F, -G:	egrep, fgrep or grep mode
	-i, +i:		case insensitive or case sensitive	
	-x, -x1, +x:	match to entire line, match to full words, or match
			    to any substring
	-A, -C, +C, +-C:approximate or exact matching
	-a, +a:		special characters represented by C-style backslash
			    notation, or not

Global Variables:

	BOOL approx;		//approximate matching in effect?
	CHAR char_map[][];	//array of character maps
	short cost_del;		//cost of deletion mismatch
	short cost_ins;		//cost of insertion mismatch
	short cost_limit;	//# of mismatches allowed for approx. matching
	short cost_sub;		//cost of substitution mismatch
	int Flags[];		//array of settings of option flags

Files:
	None:

Return Status:

	None:

*******************************************************************************/

static void
fill_in(p)
PATTERN *p;
{
	/* Parser to be used for this pattern.  Affected by -E, -F, -G */
	p->parser = parser;
	/* Character map to be used.  If -i is in effect, char_map[1] is used,
	 * which maps lower case letters into upper case, so that matches
	 * are case insensitive.  If -i is not in effect, char_map[0],
	 * the identity map, is used.
	 */
	p->cmap = char_map[Flags['i']];
	/* Full line or full word(s) matching?  Affected by -x, -x1, or +x */
	p->xflag = Flags['x'];
	/* Approximate matching parameters. Affected by -A and {-|+|+-|-+}C */
	p->cost_limit = cost_limit;
	p->cost_sub = cost_sub;
	p->cost_del = cost_del;
	p->cost_ins = cost_ins;
	p->approx = approx;
	/* Character map to be used during compilation for characters escaped
	 * by a backslash.  If -a is in effect, the special mappings
	 * in char_map[2] are used (e.g. \n to new-line, \t to tab, etc.)
	 * If -a is not in effect, the escaped character maps to itself,
	 * since char_map[0] is the identity map.
	 */
	p->bmap = char_map[2*Flags['a']];
}

/******************************************************************************

Routine:
	getoption()

Syntax:
	static int getoption(argc, argv, opts)

	int argc;	//number of arguments
	char *argv[];	//array of arguments
	char *opts;	//string of legal option letters.  Options that
			//  take an argument are followed by a single colon (':').

Description:

	Reads in the next command line option.  This is an enhanced version
	of the UN*X routine getopt(), which parses options.

Global Variables:

	char *opt_arg;	// Argument of option, if any. (Output)
	int opt_ind;	// Current index into argv[] each call.  (Modified)
	char optsign;	// '-', '+', or '0', if option letter was preceded by
			//	'-', '+', or '+-' (or '-+') respectively.  (Output)
	int r_reg;	// Optional integer after option letter for options that
			//	do not take an argument. (Output)

Files:
	None:

Return Status:

	EOF	// No more options
	'?'	// Illegal option (not in "opts")
	'0'	// Numerical option, (e.g. -5)
	Other	// Otherwise, returns option letter (e.g. 'i').

*******************************************************************************/

static int
getoption(argc, argv, opts)
int	argc;
char	**argv;
char	*opts;
{
	static int sp = 1;
	register int c;
	register char *cp;

	r_reg = 0;
	/* If sp == 1, we are at the beginning of an argument.
	 * sp will be greater than 1 only if we are in the middle
	 * of processing multiple option letters that are run
	 * together in a single argument, (e.g. -vni).
	 */
	if (sp == 1) {
		/* We must check the next argument */
		/* If we have no more arguments, return EOF. */
		if (opt_ind >= argc) {
			return(EOF);
		}
		/* If the next argument does not begin with a '-' or '+',
		 * or if it is one character in length, return EOF, as
		 * we have exhausted the options.  If the argument is "--",
		 * skip over it and return EOF, as this means that there
		 * are no more options.
		 */
		/* set optsign to the first character. */
		optsign = argv[opt_ind][0];
		if ((optsign != '-' && optsign != '+') ||
		    argv[opt_ind][1] == '\0') {
			return(EOF);
		} else if (strcmp(argv[opt_ind], "--") == 0) {
			opt_ind++;
			return(EOF);
		}
		/* Set c to the second character of the argument. */
		c = argv[opt_ind][sp];
		/* If the first two characters are "-+" or "+-",
		 * reset optsign to '0', and reset c to the next
		 * (third) character.
		 */
		if ((c == '+' || c == '-') && c != optsign) {
			optsign = '0';
			c = argv[opt_ind][++sp];
		}
		/* If the next character is a digit, we set opt_arg
		 * to the remaining string (presumably all digits),
		 * and return '0' (ASCII character.)
		 * sp is reset to 1 and opt_ind is incremented to prepare
		 * for the next argument on the next call.
		 */
		if (isdigit(c)) {
			opt_arg = &argv[opt_ind++][sp];
			sp = 1;
			return('0');
		}
	}
	/* Get the next character in the argument,
	 * presumably an option letter.
	 */
	c = argv[opt_ind][sp];
	/* Check option for legality.  If it is not in "opts",
	 * skip over it, and return '?'.
	 */
	if (c == ':' || (cp=strchr(opts, c)) == 0) {
		errmsg2("%s: illegal option -%c or +%c\n", c, c);
		if (argv[opt_ind][++sp] == '\0') {
			/* No more characters in this argument.
			 * Advance to next argument.
			 */
			opt_ind++;
			sp = 1;
		}
		return('?');
	}
	/* The option letter is legal.  See if it takes an argument. */
	if (*++cp == ':') {
		/* The option letter takes a (required) argument. */
		if (argv[opt_ind][sp+1] != '\0') {
			/* We are not at the end of the current argument.
			 * Interpret the remaining characters of the
			 * argument string as the argument for the option
			 * (e.g. -wdelimiter), and advance opt_ind past
			 * the argument.
			 */
			opt_arg = &argv[opt_ind++][sp+1];
		} else if (++opt_ind >= argc) {
			/* We were at the end of the current argument,
			 * and there are no more arguments.
			 * This is an error, as an argument is required.
			 */
			errmsg("%s: %c option requires an argument\n", c);
			sp = 1;
			return('?');
		} else {
			/* Set the argument, and advance past it. */
			opt_arg = argv[opt_ind++];
		}
		/* On the next call, we will begin the next argument. */
		sp = 1;
	} else {
		/* The option letter does not take an argument.
		 * Advance to the next character
		 */
		char k;
		++sp;
		k = argv[opt_ind][sp];
		/* cgrep supports an optional numerical modifier.
		 * If the next character is a digit, interpret this digit
		 * and any contiguous subsequent digits as a numerical
		 * modifier and set r_reg to the integer value.
		 */
		if (isdigit(k)) {
			r_reg = 0;
			while (isdigit(k)) {
				r_reg *= 10;
				r_reg += (k - '0');
				k = argv[opt_ind][++sp];
			}
		}
		/* If the option letter was the last character of the
		 * argument, advance to the beginning of the next command
		 * line argument.
		 */
		if (k == '\0') {
			sp = 1;
			opt_ind++;
		}
		/* The option does not take an argument. */
		opt_arg = 0;
	}
	/* Return the option letter */
	return(c);
}

/******************************************************************************

Routine:
	re_error()

Syntax:
	void re_error(s)

	char *s;	// Pointer to error message to output.

Description:

	Reports an error to standard error
	and exits with an exit value of 2 (error).

Global Variables:

	char *Progname;	// Name of this program (for error message)

Files:
	None:

Return Status:

	Does not return.

*******************************************************************************/


void
re_error(s)
char *s;
{
	(void)fprintf(stderr, "%s: %s\n", Progname, s);
	exit(2);
}
