	static char sccsid[]="%Z%	file.c	%E% %U%";
/******************************************************************************

                ###############################################
                # Copyright 1987-2000 Lucent Technologies Inc #
		# Author:  Bill Tanenbaum		      #
                ###############################################

Module:	file.c	[%P%]

Description:

	The main routines for searching a text file for matches to
	patterns.

Functions:
	exec_file()
	exec_buffer()
	next_buffer()
	init_file()
	read_file()


*******************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "re.h"


static int next_buffer();
static void exec_buffer(), init_file(), read_file();


/******************************************************************************

Routine:
	exec_file()

Syntax:
	void exec_file(file, pat, beg_delim, end_delim, tpat, cpat)

	char *file;		// Path of text file to be processed (0 if stdin)
				//	If -M specified, path of the first text file
	PATTERN *pat;		// Pattern to be matched
	PATTERN *beg_delim;	// Window beginning delimiter
	PATTERN *end_delim;	// Window end delimiter
	PATTERN *tpat;		// Trail pattern (-t, +t, or -T)
	PATTERN *cpat;		// Combined pattern and trail pattern
				//	(initially, a copy of tpat)

Description:

	Processes one text file (or, if -M was specified, all the text files).

Global Variables:

	char *Filename;	// File name for use in output messages
	int Flags[];	// Command line option settings
	BOOL L_done;	// Are we no longer looking for matches, because -l or -l1
			//	was specified, and we found a match?
	int Nsucc;	// 2, if an error has occurred, otherwise 1 if a match
			// 	has occurred, and 0 otherwise.  NOTE: The -s1 and
			//	-V options modify this somewhat.
	int Tln;	// Count of lines for -c or -c1 options

Files:
	Argument "file" is the text file to be processed.

Return Status:

	None

*******************************************************************************/

void
exec_file(file, pat, beg_delim, end_delim, tpat, cpat)
char *file;
PATTERN *pat, *beg_delim, *end_delim, *tpat, *cpat;
{
	int	file_desc;

	/* Open the file and do initialization */
	if (file == 0) {
		file_desc = 0; /* mark as stdin */
	} else if ( (file_desc = vpopen(file, O_RDONLY)) < 0) { 
		if (Flags['s'] < 2) {
			Nsucc = 2;
		}
		if (!Flags['s']) {
			errmsg("%s: can't open %s\n", file);
		}
		return;
	}
	init_file(pat);
	/* Now process the text */
	exec_buffer(file, file_desc, pat, beg_delim, end_delim, tpat, cpat);
	/* Close the file */
	(void)close(file_desc);

	/* Perform any final printout due to -l# or -c# options.  Note that
	 * stored value of Flags[] is one more than the specified number
	 */
	/* If -l or -l2 was specified, print file name if appropriate */
	if ((L_done && Flags['l'] == 1) || 	/* -l */
	    (!L_done && Flags['l'] == 3)) {	/* -l2 */
		/* Print due to -l or -l2 option. */
		(void)fprintf(stdout, "%s\n", Filename);
	}

	/* If -c or -c1 was specified, print file name if appropriate, and
	 * print line count.
	 */
	if (Flags['c'] == 1 || Flags['c'] && Tln) {
		if (!Flags['h']) {
			(void)fprintf(stdout, "%s:", Filename);
		}
		(void)fprintf(stdout, "%ld\n", Tln);
	}
}

/******************************************************************************

Routine:
	exec_buffer()

Syntax:
	static void exec_buffer(file, file_desc, pat, beg_delim, end_delim, tpat, cpat)

	char *file;		// Path of text file to be processed (0 if stdin)
				//	If -M specified, path of the first text file
	int file_desc;		// File descriptor of "file".
	PATTERN *pat;		// Pattern to be matched
	PATTERN *beg_delim;	// Window beginning delimiter
	PATTERN *end_delim;	// Window end delimiter
	PATTERN *tpat;		// Trail pattern (-t, +t, or -T)
	PATTERN *cpat;		// Combined pattern and trail pattern
				//	(initially, a copy of tpat)

Description:

	Called by exec_file to process one text file (or, if -M was specified,
	all the text files).

Global Variables:

	CHAR *Boffset;	// Pointer to beginning of file in text buffer
			//	(used only for offsets for -b and -b1 options).
			//	If the beginning of the file is no longer
			//	in the text buffer, "Boffset" will point to a
			//	location before the beginning of the buffer!
	CHAR *Buffend;	// Pointer to last character in text buffer.
	CHAR *Buffer;	// Pointer to first character in text buffer.
	BOOL Check_truncated;// Does running off the beginning of the buffer in a
			//	delimiter search mean that output has been truncated?
	BOOL File_more;	// Is some of the file still not read in?
	int Flags[];	// Command line option settings
	BOOL L_done;	// Are we no longer looking for matches, because -l or -l1
			//	was specified, and we found a match?
	CHAR *Last_counted;// Pointer to character (in text buffer) that was
			//	most recently printed or counted.
	CHAR *Last_NL;	// Pointer to last new-line character in buffer.
	CHAR *Last_printed;// Pointer to character (in text buffer) that was
			//	most recently printed. 
	BOOL M_done;	// Are we no longer looking for pattern matches?
	int Maxbufsize;	// Size of text buffer, in bytes.
	CHAR *Mline_mark;// Pointer to a new-line character near the end of the
			//	text buffer.  If "Match_pattern" must fit
			//	on a single line, Mline_mark is equal to "Last_NL".
			//	Otherwise, Mline_mark points to a new-line
			//	character sufficiently before "Last_NL" so
			//	that any multiline match extending after
			//	Last_NL must begin at or after Mline_mark.
	CHAR *Refill_mark;// If a file does not fit into the text buffer all at once,
			//	"Refill_mark" points to the beginning of the
			//	portion of the text in the buffer that will be kept
			//	in the buffer for processing on the next pass.

Files:
	Argument "file" is the text file to be processed.

Return Status:

	None

*******************************************************************************/

static void
exec_buffer(file, file_desc, pat, beg_delim, end_delim, tpat, cpat)
char *file;
int file_desc;
PATTERN *pat, *beg_delim, *end_delim, *tpat, *cpat;
{
	int offset;
	int savechars = 0; /* # of characters saved in buffer from previous pass */
	int savematches = 0; /* # of matches saved in buffer from previous pass */
	register CGWINDOW *w;
	register CGWINDOW *a;
	CHAR *begin = Buffer;

	/* Is there more of the file to read? */
	while (File_more) {
		/* "savechars" characters have been saved at the beginning of
		 * the buffer from the previous pass (if any).  Fill up the
		 * rest of the buffer.
		 */
		read_file(file, file_desc, Buffend, Maxbufsize-savechars);

		/* Look for matches to "pat". However, if M_done is set,
		 * we have no need to look for more matches.   M_done is
		 * typically set because -N, -l, or -l1 option was used,
		 * and further matches would have no effect.
		 */ 
		if (M_done) {
			pat->nmatches = savematches;
		} else {
			pat->nmatches = pattern_match(pat, begin, Last_NL, savematches);
			if (pat->done) {
				M_done = YES;
			}
		}
		/* Perform windowing on any matches found */
		windows(pat, beg_delim, end_delim, Buffer, File_more ? Buffend+1 : Last_NL);

		/* Perform processing for -t, +t, or -T options. */
		trail_patterns(begin, Last_NL, pat, beg_delim, end_delim, tpat, cpat, savematches);
		/* Determine how much of the buffer needs to be kept for the
		 * next pass.  The global parameter, Refill_mark, will be set
		 * to the start of the text that needs to be kept.
		 */
		savematches = next_buffer(pat, beg_delim, end_delim);

		/* Output the matched text and the surrounding windows.
		 * -V and -H flags need special attention.
		 */
		if (Flags['V']) {
			/* -V case.  Output the text OUTSIDE the windows. */
			BOOL high = (BOOL)Flags['H'];
			if (pat->nmatches) {
				w = pat->matches;
				do_output (Buffer, w->wbegin, high);
				for (++w; w < pat->matches+pat->nmatches; w++) {
					do_output ((w-1)->wend, w->wbegin, high);
				}
				do_output ((w-1)->wend, Refill_mark, high);
			} else {
				do_output (Buffer, Refill_mark, high);
			}
		} else if (Flags['H']) {
			/* -H or -H1 case.  Output windows with highlighting
			 * of the match itself, but not the surrounding
			 * windows.
			 */
			for (w = pat->matches; w < pat->matches+pat->nmatches; w++) {
				do_output (w->wbegin, w->begin, NO);
				do_output (w->begin, w->end, YES);
				do_output (w->end, w->wend, NO);
			}
		} else {
			/* Normal case.  Output windows with no highlighting. */
			for (w = pat->matches; w < pat->matches+pat->nmatches; w++) {
				do_output (w->wbegin, w->wend, NO);
			}
		}
		/* If there is more of the file to be read in, we need to
		 * prepare for the read.
		 */
		if (File_more) {
			/* If we need look for no more matches because -l or
			 * -l1 was used and a match has been found (L_done set),
			 * we need no further processing of this file.
			 * If we need look for no more matches because -N was
			 * was used and the specified number of matches was
			 * found (M_done set), we need no further processing of
			 * this file, unless:
			 * 	1) Some matches are saved for the next pass
			 * 	OR
			 * 	2) -V is on (need to print text OUTSIDE windows)
			 * 	OR
			 * 	3) -t, +t, or -T was used (need to keep
			 * 	searching for trail patterns).
			 */
			if (L_done || (M_done && !savematches &&
			    !Flags['V'] && !Flags['t'])) {
				/* Don't need to check rest of file */
				break;
			}
			/* Now that the buffer will no longer contain the
			 * beginning of the file, we may have to check that the
			 * beginning of the next window is not truncated due
			 * to the buffer size.  However, if -o was not used,
			 * and the last match output on the previous pass
			 * extends to the refill mark, nothing can be lost.
			 */
			if (!Flags['o'] && pat->nmatches > 0 &&
			    pat->matches[pat->nmatches-1].wend == Refill_mark) {
				Check_truncated = NO;
			} else if (beg_delim->wcount) {
				Check_truncated = YES;
			}

			/* calculate # of characters to be saved for next pass */
			savechars = Buffend - Refill_mark;

			/* calculate offset for pointers */
			offset = Refill_mark - Buffer;

			/* This call to do_output() will not output anything,
			 * since it has zero range.  However, it is needed
			 * to maintain the line count for the -n option.
			 */
			do_output(Refill_mark, Refill_mark, NO);

			/* Move any text to be saved for the next pass
			 * up to the beginning of the buffer.
			 * Adjust any pointers that point into the buffer
			 * by the move offset. 
			 * The values of Refill_mark and Mline_mark are
			 * reset on the next pass, so they need not be adjusted.
			 * Also, Buffer always points to the beginning
			 * of the buffer, so it must not be adjusted.
			 */
			Last_counted -= offset;
			Last_printed -= offset;
			Last_printed = MAX(Buffer-1, Last_printed);
			(void)memcpy(Buffer+1, Refill_mark+1, savechars);
			Boffset -= offset;
			Buffend -= offset;
			for (a = pat->matches, w = a + pat->nmatches;
			   a < pat->matches+savematches; a++, w++) {
				/* Note that we wipe out the end of window
				 * information for the saved matches by
				 * resetting "wend" to 0.  This windowing
				 * will be done again on the next pass.
				 */
				a->wbegin = w->wbegin - offset;
				a->begin = w->begin - offset;
				a->end = w->end - offset;
				a->wend = 0;
				a->prev_matches = w->prev_matches;
			}
			/* Any matches beginning before Mline_mark have
			 * already been found, so start looking at
			 * Mline_mark for pattern matches on the next pass.
			 */
			begin = Mline_mark - offset;
		}
	} /* end while (File_more) */ 
}

/******************************************************************************

Routine:
	next_buffer()

Syntax:
	static void next_buffer(pat, beg_delim, end_delim)

	PATTERN *pat;		// Pattern to be matched
	PATTERN *beg_delim;	// Window beginning delimiter
	PATTERN *end_delim;	// Window end delimiter

Description:

	Called by exec_buffer() to determine how much of the buffer contents
	needs to be saved for the next pass, and to handle all the complexities
	of having windows split across buffers.

	Refill_mark will be set to point to the beginning of the text that needs
	to be saved for the next pass.

Global Variables:

	CHAR *Buffend;	// Pointer to last character in text buffer.
	CHAR *Buffer;	// Pointer to first character in text buffer.
	BOOL File_more;	// Is some of the file still not read in?
	int Flags[];	// Command line option settings
	CHAR *Last_NL;	// Pointer to last new-line character in buffer.
	BOOL M_done;	// Are we no longer looking for pattern matches?
	CHAR *Mline_mark;// Pointer to a new-line character near the end of the
			//	text buffer.  If "Match_pattern" must fit
			//	on a single line, Mline_mark is equal to "Last_NL".
			//	Otherwise, Mline_mark points to a new-line
			//	character sufficiently before "Last_NL" so
			//	that any multiline match extending after
			//	Last_NL must begin at or after Mline_mark.
	CHAR *Refill_mark;// If a file does not fit into the text buffer all at once,
			//	"Refill_mark" points to the beginning of the
			//	portion of the text in the buffer that will be kept
			//	in the buffer for processing on the next pass.
	BOOL Window_truncated;// Has a window possibly been lost or truncated
			//	because the buffer size is too small?

Files:
	None

Return Status:

	Number of matches to save for next pass

*******************************************************************************/

static int
next_buffer(pat, beg_delim, end_delim)
PATTERN *pat, *beg_delim, *end_delim;
{
	register int savematches = 0;
	register CGWINDOW *w;
	register CGWINDOW *a;

	/* If this is the last buffer of the file, we just set the refill
	 * mark to the end of the buffer, since this will be the last pass.
	 */
	if (!File_more) {
		Refill_mark = Last_NL;
		return(0);
	}

	/* We set the refill mark back far enough so that we will not truncate
	 * the beginning of a future window, given sufficient room in the buffer.
	 * We do this by searching for the beginning delimiters, if necessary.
	 */
	if (M_done && !Flags['t']) {
		/* Case 1: We are no longer looking for  patterns or trail
		 * patterns, so set the refill mark to the end of the 
		 * last full line in the buffer.
		 */
		Refill_mark = Last_NL;
	} else if (!pat->nmatches) {
		/* Case 2: We have no matches.
		 * Therefore, we need to search for the begin delimiter
		 * all the way back to the beginning of the buffer.
		 */
		Refill_mark = window_limit(beg_delim, Buffer, Mline_mark, 0);
	} else {
		/* Case 3: We have matches.
		 * Therefore, we need to search for the begin delimiter
		 * only back to the end of the last window, or, if -o was
		 * used, back to the beginning of the last window.
		 */
		CHAR *e;
		if (Flags['o']) {
			e = pat->matches[pat->nmatches-1].wbegin;
		} else {
			e = pat->matches[pat->nmatches-1].wend;

		}
		if (e > Mline_mark) {
			Refill_mark = Mline_mark;
		} else {
			Refill_mark = window_limit(beg_delim, e, Mline_mark, 0);
		}
	}
	/*
	 * If there are any matches in this buffer, we may need to adjust
	 * the Refill_mark, and/or postpone processing some windows to the
	 * next pass.
	 */ 
	if (pat->nmatches) {
		/* 
		 * To avoid truncating the end of a window, windows running
		 * off the end of the buffer need some processing on the next
		 * pass.  Also, in the normal (-o option not used) case,
		 * we process any text past the Refill mark on the next pass,
		 * even for windows that do not run off the end of the buffer.
		 * This is the simplest way to ensure that each line in a
		 * window will be output only once.
		 *
		 * So, in more detail:
		 * Normal (no -o case):
		 * All windows beginning at or after the current Refill mark
		 * will be postponed to the next pass.
		 * Any window straddling the current Refill mark (there can be
		 * at most one such window) will be split into two windows
		 * at the Refill mark, which may need adjustment before the
		 * split.  The second of the two windows will be postponed
		 * to the next pass.
		 *
		 * (-o case):
		 * All windows that run off the end of the buffer will be
		 * postponed to the next pass, except possibly the first
		 * such window.  If any of these windows straddle the
		 * Refill mark, the Refill mark will be set back to the
		 * beginning of the window.
		 * Then, the first such window will be postponed
		 * if it begins at or after the Refill mark.  If it
		 * straddles the Refill mark, it will be split into two
		 * windows at the Refill mark, which may need adjustment.
		 * The second of the two windows will be postponed
		 * to the next pass.
		 */
		CHAR *checkpoint = (Flags['o'] ? Last_NL : Refill_mark);
		/* Loop through the windows in reverse order */
		for (w = pat->matches+pat->nmatches-1;
		    w >= pat->matches && w->wend > checkpoint; w--) {
			CHAR *search_end_delim;
			int wc;

			/* We have a postponement.  */
			savematches++;

			/* If window is completely past Refill_mark,
			 * postpone the entire window.
			 */
			if (w->wbegin >= Refill_mark) { 
				continue;
			}
			/* If the previous window also goes off the
			 * end, postpone this entire window.
			 * This only can happen in the -o case.
			 */
			if (w != pat->matches && (w-1)->wend > Last_NL) { 
				Refill_mark = MIN(Refill_mark, w->wbegin);
				continue;
			}
			/* We split this window that straddles the
			 * Refill mark.  However, we can't necessarily
			 * split it simply at the Refill mark, since there
			 * may have been matches to the end delimiter before
			 * the Refill mark.  So, we have several cases
			 * we must consider.  First, find the point at which
			 * the search for the end delimiter must begin.
			 */
			search_end_delim = w->end;

			if (end_delim->incl_match && search_end_delim > Buffer) {
				search_end_delim = memchr_back(search_end_delim-1, NL);
				search_end_delim = MAX(w->begin, search_end_delim);
			}

			/* Number of delimiter matches needed to terminate window. */
			wc = end_delim->wcount - w->prev_matches;

			if (search_end_delim > Buffer) {
				/* Case 1: The delimiter search does not begin
				 * at the beginning of the buffer.  In this case,
				 * we can split the window at the beginning of
				 * the delimiter search.
				 */
				Refill_mark = MIN(Refill_mark, search_end_delim);
			} else if (wc == 1) {
				/*
				 * Case 2: We are searching for a single match
				 * to the end delimiter.  We know we did not find
				 * it before the Refill mark.  Therefore, we can
				 * split the window at the Refill mark, unless
				 * the end delimiter can span multiple lines,
				 * in which case we must back up to make sure
				 * we don't miss a delimiter.
				 */
				int i = end_delim->mult_lines;
				while (i > 0 && Refill_mark > Buffer) {
					Refill_mark = memchr_back(Refill_mark-1, NL);
					i--;
				}
			} else {
				int i;
				int mat;
				/* Case 3: We were searching for more than
				 * one match to the end delimiter.  We didn't
				 * find all that we needed.  So, find as many
				 * as we can.
				 */
				/* Search for matches to end_delim */
				end_delim->N_start = 0;
				end_delim->N_stop = wc;
				mat = pattern_match(end_delim, search_end_delim, Refill_mark, 0);
				if (mat) {
					/* There are matches.  Increment
					 * the count.
					 */
					w->prev_matches += mat;
					/* 
					 * If the end delimiter can span
					 * multiple lines, back up the
					 * Refill mark to avoid missing
					 * an end delimiter on the next pass.
					 */
					i = end_delim->mult_lines;
					while (i > 0 && Refill_mark > end_delim->matches[0].end) {
						Refill_mark = memchr_back(Refill_mark-1, NL);
						i--;
					}
				} else {
					/* No match. Treat like Case 2. */
					i = end_delim->mult_lines;
					while (i > 0 && Refill_mark > Buffer) {
						Refill_mark = memchr_back(Refill_mark-1, NL);
						i--;
					}
				}
			}
			/* If the Refill_mark still points to the beginning of
			 * the buffer, the window can't be split.
			 */
			if (Refill_mark == Buffer) {
				break;
			}

			/* OK. Now split the window at the Refill_mark.
			 * But first, copy the windows upward in the
			 * matches buffer to make room.  w itself
			 * will be duplicated.
			 */

			match_alloc(pat);
			for (a = pat->matches+pat->nmatches; a > w ; a--) {
				*a = *(a-1);
			}
			a = w + 1;
			/* OK. Now split them. w->wbegin and a->wend
			 * do not need adjusting.
			 */
			pat->nmatches++;
			w->begin = MIN(w->begin, Refill_mark);
			w->end = w->wend = a->wbegin = Refill_mark;
			a->begin = MAX(a->begin, Refill_mark);
			a->end = MAX(a->end, a->begin);

			/* No more windows to split or postpone. */
			break;
		}

		/* Decrease the match count by the number of postponed matches,
		 */
		pat->nmatches -= savematches;

		/* Set w to point to the earliest postponed window */
		w = pat->matches + pat->nmatches;
	}
	/* We are probably all done.  However, if the Refill_mark is
	 * at the beginning of the buffer, we will go into an infinite loop
	 * unless we advance it somehow!
	 */
	if (Refill_mark == Buffer) {
		/* Check if the first line of the buffer is in a window. */
		if (Last_NL == Buffer) {
			/*
			 * A single line did not fit into the buffer.
			 * Toss the contents of the buffer.
			 * This may cause a window to be missed or truncated.
			 */
			Line_truncated = YES;
			Refill_mark = Mline_mark = Buffend;
		} else if (savematches && w->wbegin == Buffer) {
			/* The first line of the buffer is in a postponed
			 * window.  We must truncate the window.
			 */
			Window_truncated = YES;

			/* Advance the Refill_mark to the end of that window,
			 * but don't go beyond Mline_mark.  Don't do this
			 * if -o is on.  We still try to preserve overlap!
			 */
			if (!Flags['o']) {
				Refill_mark = MIN(w->wend, Mline_mark);
			}

			/* If Refill_mark is still at the beginning of the
			 * buffer, advance it one line.
			 * If there are no new-lines in the entire buffer,
			 * we will stop at the sentinel NL one past the end of
			 * the buffer, in which case Refill_mark must be set
			 * to the end of the buffer.
			 */
			if (Refill_mark == Buffer) {
				Refill_mark = memchr_fwd(Refill_mark+1, NL);
				if (Refill_mark > Buffend) Refill_mark = Buffend;
				Mline_mark = MAX(Refill_mark, Mline_mark);
			}
			/* Truncate the end of the window at the Refill_mark,
			 * and don't postpone it.  Don't truncate if -o is on.
			 */
			if (Flags['o']) {
				w->wend = MIN(w->wend, Last_NL);
			} else {
				w->wend = Refill_mark;
				w->end = MIN(w->end, w->wend);
				w->begin = MIN(w->begin, w->wend);
			}
			/* Process the truncated window on this pass, not
			 * on the next pass.
			 */
			pat->nmatches++;
			savematches--;
			/* In the -o case, all subsequent postponed windows
			 * that go back to the beginning of the buffer
			 * must be processed on this pass.
			 */
			if (Flags['o']) {
				int i;
				for (i = 0; i < savematches; i++) {
					/* Get next window */
					w++;
					if (w->wbegin == Buffer) {
						w->wend = MIN(w->wend, Last_NL);
					} else {
						break;
					}
				}
				pat->nmatches += i;
				savematches -= i;
			}
		} else {
			/* The first line of the buffer is not in a window.
			 * Advance the Refill_mark.  The lines passed over
			 * will be printed by exec_buff() if -V is on.
			 * Advancing by one line at a time is very slow.
			 * Therefore, we advance by about half the buffer,
			 * but not past the beginning of the first postponed
			 * window, if any.
			 */
			/* Advance half the buffer */
			Refill_mark = Buffer + ((Buffend - Buffer)/2); 
			/* Back up to beginning of line. */
			Refill_mark = memchr_back(Refill_mark-1, NL);
			/* If now at beginning of buffer, advance one line */
			if (Refill_mark == Buffer) Refill_mark = memchr_fwd(Refill_mark+1, NL);
			/* Make sure we are not off the end of the buffer. */
			if (Refill_mark > Buffend) Refill_mark = Buffend;
			/* Don't go past first postponed window, if any. */
			if (savematches && (Refill_mark > w->wbegin)) {
				Refill_mark = w->wbegin;
			}
			Mline_mark = MAX(Refill_mark, Mline_mark);
		}
	}
	return(savematches);
}

/******************************************************************************

Routine:
	init_file()

Syntax:
	static void init_file(pat)

	PATTERN *pat;		// Pattern to be matched

Description:

	Called by exec_file() to initialize parameters for processing the file.

Global Variables:

	BOOL Any_printed;	// Has any output been printed?
	CHAR *Boffset;		// Pointer to beginning of file in text buffer
				//   (used only for offsets for -b and -b1 options).
				//   If the beginning of the file is no longer
				//   in the text buffer, "Boffset" will point to a
				//   location before the beginning of the buffer!
	CHAR *Buffend;		// Pointer to last character in text buffer.
	CHAR *Buffer;		// Pointer to first character in text buffer.
	BOOL Check_truncated;	// Does running off the beginning of the buffer in a
				//   delimiter search mean that output
				//   has been truncated?
	BOOL File_more;		// Is some of the file still not read in?
	BOOL L_done;		// Are we no longer looking for matches, because
				//    -l or -l1 was specified, and we found a match?
	CHAR *Last_counted;	// Pointer to character (in text buffer) that was
				//    most recently printed or counted.
	CHAR *Last_printed;	// Pointer to character (in text buffer) that was
				//    most recently printed. 
	int Line_number;	// Line number for -n
	BOOL M_done;		// Are we no longer looking for pattern matches?
	int Tln;		// Count of lines for -c or -c1 options

Files:
	None

Return Status:

	None

*******************************************************************************/


static void
init_file(pat)
PATTERN *pat;
{
	Check_truncated = NO;	/* Has beginning of file already been overwritten
				 * in the buffer, so we must check for truncation
				 * at the beginning of each window? NO.
				 */

	L_done = NO;		/* Are we done with matching because we used
				 * -l or -l1 and have found a match? NO.
				 */
	M_done = NO;		/* Are we done with matching for any reason
				 * (i.e. we have found the number of matches
				 * specified in the -N option or because of -l or
				 * -l1 (as in L_done above))?  NO, unless we
				 * know now that we will never have any matches.
				 */
	if (pat->procfn == re_nullexec && !pat->vfast && !pat->vflag) {
		/* We can have no matches, (and -v is not on). */
		M_done = YES;
	}
	File_more = YES;	/* Is more of file to be read? YES. */

	Tln = 0;		/* count of lines for -c or -c1 */

	Line_number = 1;	/* Line number for -n */

	/* Set buffer pointers to beginning of text buffer */
	Buffend = Boffset = Last_counted = Last_printed = Buffer;

	/* If we have printed anything from a previous file, decrement
	 * Last_printed so do_output() will print the separator (if any)
	 * before the first output.
	 */
	if (Any_printed) {
		Last_printed--;
	}
	pat->N_start = Nlines[1];	/* # of matches to skip (+N option) */

	pat->N_stop = Nlines[0];	/* maximum # of matches (-N option) */
}

/******************************************************************************

Routine:
	read_file()

Syntax:
	static void read_file(file, file_desc, buf, bufsize)

	char *file;	// Name of text file being read
	int file_desc;	// File descriptor of "file"
	CHAR *buf;	// Pointer to beginning of memory to read in text
	int bufsize;	// Maximum number of bytes to be read in

Description:

	Called by exec_buffer to fill the text buffer by reading a text file.
	A sentinel new-line character will be inserted into the buffer
	after the last character read.

Global Variables:

	CHAR *Buffend;	// Pointer to last character in text buffer.
	CHAR *Buffer;	// Pointer to first character in text buffer.
	BOOL File_more;	// Is some of the file still not read in?
	int Flags[];	// Command line option settings
	CHAR *Last_NL;	// Pointer to last new-line character read into the buffer,
			//	(unless	the entire file has been read in and the
			//	last line is not terminated by a new-line, in
			//	which case "Last_NL" will point to a sentinel
			//	new-line after the last line.
	BOOL M_done;	// Are we no longer looking for pattern matches?
	int Margc;	// File count for -M (merge files) option.
	char **Margv;	// Array of file names for -M option.
	CHAR *Mline_mark;// Pointer to a new-line character near the end of the
			//	text buffer.  If "Match_pattern" must fit
			//	on a single line, Mline_mark is equal to "Last_NL".
			//	Otherwise, Mline_mark points to a new-line
			//	character "Mlines" lines before "Last_NL", so
			//	that any multiline match extending after
			//	Last_NL must begin at or after Mline_mark.
	long Mlines;	// The maximum number of new-line characters that a single
			//	instance of the text matching "Match_pattern"
			//	may contain, not including the first and last
			//	characters of the match.  (e.g. Mlines == 0,
			//	if the pattern must be on a single line).
	BOOL Window_truncated;// Has a window possibly been lost or truncated
			//	because the buffer size is too small?

Files:
	The file specified by "file" and "file_desc" is read.

Return Status:

	None

*******************************************************************************/

static void
read_file(file, file_desc, buf, bufsize)
char *file;
int file_desc;
CHAR *buf;
register int	bufsize;	/* free space in data buffer */
{
	register int	bytes;	  /* bytes read in current buffer */

	Buffend = buf;

	while (1) {
		/* Keep reading until the buffer is full, or until there
		 * is nothing left to read.
		 */
		do {
			bytes = read(file_desc, (char *)Buffend+1, (unsigned int)bufsize);
			if (bytes < 0) {
				/* can't read any more of file */
				bytes = 0;
			}
			Buffend += bytes;
			bufsize -= bytes;
		} while (bytes > 0 && bufsize > 0);

retry:
		/* We have either filled the buffer, or read the entire file.
		 * If we have filled the buffer, or if -M was not specified,
		 * or if there are no more files to merge, stop reading.
		 */
		if (bufsize == 0 || !Margc) {
			break;
		}
		/* There is more room in buffer, and we are merging
		 * (-M specified) with the next file.  First, close this file.
		 */
		(void)close(file_desc);
		/* Decrement the remaining file count, and get the path to the
		 * next file.
		 */
		Margc--;
		file = *Margv++;
		/* Open the next file */
		if ((file_desc = vpopen(file, O_RDONLY)) < 0) { 
			/* Can't open the file.  Unless -s1 was specified,
			 * the exit code of cgrep is set to '2' (error).
			 */
			if (Flags['s'] < 2) {
				Nsucc = 2;
			}
			/* Unless -s or -s1 was specified, give an error
			 * message (non-fatal).
			 */
			if (!Flags['s']) {
				errmsg("%s: can't open %s\n", file);
			}
			/* Try the next file, if any. */
			goto retry;
		}
	}

	if (bufsize) {
		/* Still space in the buffer, so we have read entire file */
		File_more = NO;
		/* Set Last_NL to point to last character in buffer. If this
		 * is not a new-line, we increment Last_NL to point to the
		 * sentinel new-line character we will later insert.
		 */
		Last_NL = Buffend;
		if (*Last_NL != NL) {
			Last_NL = Buffend+1;
		}
		Mline_mark = Last_NL;
	} else {
		/* We filled entire buffer, so there may be more yet to read */
		File_more = YES;
		/* Set Last_NL to point to last new-line read in. */
		Mline_mark = Last_NL = memchr_back(Buffend, NL);
		if (Mlines && !M_done) {
			/* The pattern can span multiple lines, and we are still
			 * looking for matches.  We move Mline_mark back so that
			 * we do not miss a multiline match due to it not being
			 * completely in the buffer.
			 */
			register int i;
			for (i = 0; i < Mlines && Mline_mark > Buffer; i++) {
				Mline_mark = memchr_back(Mline_mark-1, NL);
			}
		}
		if (Mlines && Mline_mark == Buffer) {
			/* We may miss a match because the pattern may span
			 * multiple lines, and enough lines do not fit into
			 * the buffer to be sure that we will not miss a
			 * match.  Since this case is a very low runner,
			 * and since checking if we really miss a match
			 * in this case is difficult, we simply set
			 * "Window_truncated" because of the possibility
			 * of a missed match.
			 */
			Mline_truncated = YES;
		}
	}
		/* Note: Buffend is last good char */
	*(Buffend+1) = NL;	/* Sentinel new-line character */
	return;
} /* end of function read_file */
